import java.util.*;

public class VierGewinntDistributedDB{
    private final static boolean OUTPUTDBSTATS = VP.OUTPUTDBSTATS;

    private final static int DBCOUNT = VP.CONNECTIONSTRINGS.length;
    private final VierGewinntDB[] databases = new VierGewinntDB[DBCOUNT];

    //Load balance stuff
    private final static int LBGRANULARITY = 100;
    private final static int INITIALLBCOUNT = DBCOUNT*LBGRANULARITY;
    private final long[] timeUsed = new long[DBCOUNT];
    private final int[] lbPriorities = new int[DBCOUNT];
    private final int[] lbLastPriorities = new int[DBCOUNT];
    private int[] lbDistribution;
    private int lbDistributionCount;

    //HashSet to cache the last entries, because they are often similar
    private final static int MAXHASHSIZE = VP.DBCACHESIZE;
    private final HashSet hash = new HashSet(MAXHASHSIZE);
    //hashList has the same variables as hash. It is needed to know which
    //one is the oldes entry.
    private final LinkedList hashList = new LinkedList();

    private WatchDog watchDog = new WatchDog();

    /**
       Terminates the program if the WatchDog isn't toggled for
       some time;
     */
    private class WatchDog extends Thread{
	private VierGewinntDB lastDB;
	private boolean timeout = false;

	public WatchDog(){
	    super();
	    setPriority(Thread.MIN_PRIORITY);
	    setDaemon(true); //application terminates if run hasn't terminated
	    start();
	}

	public void run(){
	    boolean finished = false;
	    while(true){
		try{
		    sleep(60*60*1000); //1 hour
		   if(timeout){
			System.err.println("Database timeout for:"+lastDB);
			System.exit(-1);
		    }else timeout = true;
		}catch(InterruptedException e){
		}
	    }
	}

	public void toggle(VierGewinntDB lastDB){
	    this.lastDB = lastDB;
	    timeout=false;
	}
    }

    public VierGewinntDistributedDB (){
	for(int i=0;i<DBCOUNT;i++){
	    System.out.println(
	       "Opening database connection: "+VP.CONNECTIONSTRINGS[i]);
	    databases[i] = new VierGewinntDB(VP.CONNECTIONSTRINGS[i]);
	    lbPriorities[i] = LBGRANULARITY;
	    lbLastPriorities[i] = LBGRANULARITY;
	}
	
	createDistributionFromPriority();
    }

    public void open(){
	for(int i=0;i<DBCOUNT;i++){
	    watchDog.toggle(databases[i]);
	    databases[i].open();
	}
    }

    public void close(){
	for(int i=0;i<DBCOUNT;i++){
	    watchDog.toggle(databases[i]);
	    databases[i].close();
	}
	hash.clear();
	hashList.clear();
    }

    public void switchToNextGeneration(){
	doLoadBalance();
	for(int i=0;i<DBCOUNT;i++){
	    watchDog.toggle(databases[i]);
	    databases[i].switchToNextGeneration();
	}
	hash.clear();
	hashList.clear();
    }
    
    private void doLoadBalance(){
	final double[] tmpPriorities = new double[DBCOUNT];
	System.arraycopy(lbPriorities,0,lbLastPriorities,0,DBCOUNT);

	long timeSum = 0;
	for(int i=0;i<timeUsed.length;i++)
	    timeSum += timeUsed[i];
	final double avgTime = (double)timeSum / (double)DBCOUNT;

	if(avgTime>2*1000){
	    int prioritySum = 0;
	    for(int i=0;i<DBCOUNT;i++)
		prioritySum += lbPriorities[i];

	    for(int i=0;i<DBCOUNT;i++){
		double timeRel = (double)timeUsed[i] / avgTime;

		//increase stability
		timeRel += (1.0 - timeRel) * 0.5;
	
		tmpPriorities[i] = ((double)lbPriorities[i]/timeRel); 
	    }

	    //normalize
	    double tmpPrioritySum = 0.0;
	    for(int i=0;i<DBCOUNT;i++)
		tmpPrioritySum += tmpPriorities[i];

	    for(int i=0;i<DBCOUNT;i++){
		int p = (int)((tmpPriorities[i]/tmpPrioritySum)
			      *(double)INITIALLBCOUNT);
		if(p<=0)
		    p = 1;
		lbPriorities[i] = p;
	    }

	    createDistributionFromPriority();
	}

	if(OUTPUTDBSTATS){
	    System.out.println("Database Statistics:");
	    System.out.println(
		 "#\tPriority(old/new)\tUsagetime (this Generation)");
	    for(int i=0;i<timeUsed.length;i++){
		System.out.println(i+"\t"+lbLastPriorities[i]+"/"+lbPriorities[i]
				   +"\t\t\t"+VP.timeToString(timeUsed[i]));
	    }
	}

	for(int i=0;i<timeUsed.length;i++)
	    timeUsed[i] = 0;
    }

    private void createDistributionFromPriority(){
	//calculate the sum of the priorities
	int prioritySum = 0;
	for(int i=0;i<DBCOUNT;i++)
	    prioritySum += lbPriorities[i];

	lbDistribution = new int[prioritySum];
	lbDistributionCount = lbDistribution.length;

	int pos = 0;
	for(int i=0;i<DBCOUNT;i++){
	    int p = lbPriorities[i];
	    for(int j=0;j<p;j++)
		lbDistribution[pos++] = i;
	}
    }

    public Iterator iterator(){
	return new DistributedDBIterator();
    }

    public void add(VierGewinntTurn t){
	if(hash.add(t)){
	    //store in hashList to memorize which is the last
	    hashList.addLast(t);
	    //find out in which database the data should be stored
	    int hashCode = t.hashCode();
	    //int dbNum = hashCode%DBCOUNT;
	    final int hashLimited = (hashCode>0) ? 
		hashCode%lbDistributionCount : (-hashCode)%lbDistributionCount;
	    final int dbNum = lbDistribution[hashLimited];
	
	    //select the active database
	    final VierGewinntDB db = databases[dbNum];
	
	    //the last life sign
	    watchDog.toggle(db);

	    //store
	    db.add(t);
	
	    //calculate time for load balancing
	    timeUsed[dbNum] += db.getTimeUsedAndReset();
	    //put the entry in the insert queue if not in the hash


	    if(hash.size()>=MAXHASHSIZE){
		//remove oldest entry
		hash.remove(hashList.removeFirst());
		    //hash.clear();
	    }
	}
    }

    public long size(){
	long size = 0;
	for(int i=0;i<DBCOUNT;i++){
		watchDog.toggle(databases[i]);
	   size +=  databases[i].size();
	}
	return size;
    }

    private class DistributedDBIterator implements Iterator{
	//the number of the iterator for the next element
	private int currentIterator = 0;

	//the number of valid elements in the iterators array
	private int nonEmptyIteratorsCount;

	//contains the iterators. At no time during the classes life circle
	//every iterator must have more elements
	//if an iterator has no more elements, it must be replaced,
	//by the last valid element of the array
	//the nonEmptyIteratorCount must be decreased
	final private Iterator[] iterators = new Iterator[DBCOUNT];

	final private int[] priorities = new int[DBCOUNT];
	private int elementsToGet;

	public DistributedDBIterator(){
	    //create a copy of the priorities
	    System.arraycopy(lbLastPriorities,0,priorities,0,DBCOUNT);
	    
	    nonEmptyIteratorsCount = 0;
	    for(int i=0;i<DBCOUNT;i++){
		Iterator it = databases[i].iterator();
		if(it.hasNext()){
		    iterators[nonEmptyIteratorsCount++] = it;
		}
	    }

	    elementsToGet = priorities[currentIterator];
	}

	public boolean hasNext(){
	    return nonEmptyIteratorsCount>0;
	}

	public Object next(){
	    Iterator it = iterators[currentIterator]; 
	    Object retVal = it.next();
	    elementsToGet--;
	    
	    if(it.hasNext()){
		if(elementsToGet<=0){
		    //switch between the databases to distribute the load
		    currentIterator=(currentIterator+1)%nonEmptyIteratorsCount;
		    elementsToGet = lbPriorities[currentIterator];
		}
	    }else{
		//replace the current Iterator in the array with the last one
		if(currentIterator == (nonEmptyIteratorsCount-1)){
		    //it's already the last element
		    nonEmptyIteratorsCount--;
		    currentIterator = nonEmptyIteratorsCount-1;
		}else{
		    iterators[currentIterator] = 
			iterators[--nonEmptyIteratorsCount];
		    priorities[currentIterator] = 
			priorities[nonEmptyIteratorsCount];
		}
		if(currentIterator>0)
		    elementsToGet = priorities[currentIterator];
	    }

	    return retVal;
	}

	public void remove(){
	    throw new RuntimeException("remove is not implemented");
	}
    }
}


