import java.util.*;
import java.text.DecimalFormat;

final public class VierGewinntTreeGeneration{
    private HashSet hash = null;
    private VierGewinntDistributedDB db = null;
    private LinkedList generationStats = new LinkedList();
    private int currentPlayer = VierGewinntTurn.SECONDPLAYER;

    //some statistics
    //number of valid turns is stored in hash.size()
    private static long firstWins = 0;
    private static long secondWins = 0;
    private static long turnCount = 0;

    private class GenerationStat{
	private long count;
	public long getCount(){return count;}
	private long time;
	public long getTime(){return time;}

	public GenerationStat(long count, long time){
	    this.count = count;
	    this.time=time;
	}
	
	public String toString(){
	    //NumberFormat nf = NumberFormat.getNumberInstance();
	    DecimalFormat df = new DecimalFormat("##0.000E00");;
	    df.setGroupingUsed(false);

	    StringBuffer b = new StringBuffer();

	    final long timeBy1k = time/1000;
	    if(timeBy1k>0)
		b.append(df.format(count/timeBy1k));
	    else b.append("n/a\t");
	    b.append('\t');
	    b.append(df.format(count)).append('\t');
	    b.append(VP.timeToString(time));
	    return b.toString();
	}
    }

    public VierGewinntTreeGeneration(){
	VierGewinntTurn t = new VierGewinntTurn();
	if(VP.USEDATABASE){
	    db = new VierGewinntDistributedDB();
	    db.open();
	    db.add(t);
	    db.close();
	 }else{
	     hash = new HashSet();	
	     hash.add(t);
	 }
    }

    public long calculateNextGeneration(){
	final long startTime = System.currentTimeMillis();

	if(currentPlayer==VierGewinntTurn.FIRSTPLAYER)
	    currentPlayer = VierGewinntTurn.SECONDPLAYER;
	 else currentPlayer = VierGewinntTurn.FIRSTPLAYER;

	Iterator it;
	
	if(VP.USEDATABASE){
	    it = db.iterator();
	    db.switchToNextGeneration();
	    db.open();
	}else{
	    it = hash.iterator();
	    hash = new HashSet((int)((float)hash.size()*1.5));
	}
	    

	while(it.hasNext()){
	    VierGewinntTurn rootTurn = (VierGewinntTurn)it.next();

	    int result = rootTurn.getWinner();

	    if(result==VierGewinntTurn.NOBODY){
		//these two are needed to eliminate dumb turns
		boolean hasWinner = false;
		Vector turns;
		if(VP.STOREDUMBTURNS)
		    turns = null; //don't need it
		 else turns = new Vector(VP.MAXCOLUMNS);
		
		for(int i=0;i<VP.MAXCOLUMNS;i++){
		    if(rootTurn.columnHasSpace(i)){
			turnCount++;
			VierGewinntTurn t 
			    =new VierGewinntTurn(rootTurn,currentPlayer,i);
			if(VP.OUTPUTALL)
			    System.out.println(t.toString());

			int nextRes = t.getWinner();

			switch(nextRes){
			case VierGewinntTurn.FIRSTPLAYER: 
			    firstWins++;
			    hasWinner = true;
			    storeTurn(t); //store the turn now
			    if(VP.OUTPUTWON&&!VP.OUTPUTALL)
				System.out.println(toString());
			    break;
			case VierGewinntTurn.SECONDPLAYER: 
			    secondWins++;
			    hasWinner = true;
			    storeTurn(t); //store the turn now
			    if(VP.OUTPUTWON && !VP.OUTPUTALL)
				System.out.println(toString());
			    break;
			}

			if(VP.STOREDUMBTURNS)
			    storeTurn(t);
			 else turns.add(t);
		    }//if comlunHasSpace
		}//for

		if(!VP.STOREDUMBTURNS){
		    if(!hasWinner){
			Enumeration enum = turns.elements();
			while(enum.hasMoreElements())
			    storeTurn((VierGewinntTurn) enum.nextElement());
		    }
		}
	    }//if(result==VierGewinntTurn.NOBODY)

	    if(!VP.USEDATABASE)
		it.remove(); //free the memory
	    
	}//while(it.hasNext())

	long count;
	if(VP.USEDATABASE){
  	    count = db.size();
	    db.close();
	}else{
	    count = hash.size();
	}

	final long endTime = System.currentTimeMillis();

	generationStats.add(new GenerationStat(count,endTime-startTime));

	return count;
    }

    /**
       @return false if the entry did exist before
    */
    private void storeTurn(VierGewinntTurn t){
	if(VP.USEDATABASE){
	    db.add(t);
	}else{
	    //don't store the turn if it already existed
	    if(!hash.contains(t)){
		if(VP.OUTPUTALL)
		    System.out.println(t.toString());
		hash.add(t);
	    }
	}
    }

    public String toString(){
	return getStatistics();
    }

    public String getStatistics(){
	StringBuffer out = new StringBuffer();
	out.append("Number of calculated turns: ").append(turnCount);
	long numOfValidTurns = 0;
	Iterator it = generationStats.iterator();
	while(it.hasNext())
	    numOfValidTurns += ((GenerationStat)it.next()).getCount();
	out.append("\nNumber of valid turns: ").append(numOfValidTurns);
	out.append("\nFirst player wins: ").append(firstWins);
	out.append("\nSecond player wins: ").append(secondWins);
	out.append('\n');
	out.append("Generation Stats:\n");
	out.append("#\tturns/sec\t# of turns\tcalcTime\n");
	it = generationStats.iterator();
	int i=0;
	while(it.hasNext()){
	    out.append(i+1).append('\t');
	    out.append(it.next()).append('\n');
	    i++;
	}
	out.append('\n');
	return out.toString();
    }
}



