public final class VierGewinntTurn{
    /*
       The positions of the chips in the matrix are stored in theese
       two long values. The MSB indicates the upper left position,
       and the LSB the lower right. the positions are counted from
       left to right.
     */
    private long firstPlayerChips=0;
    private long secondPlayerChips=0;

    public final static int FIRSTPLAYER = 1;
    public final static int SECONDPLAYER = 2;
    public final static int DRAW = 3;
    public final static int NOBODY = 4;

    private final static long COLMASK = 0x0101010101010101l;

    //The number of bits in a byte
    private final static int[] BITSPERSHORT = new int[65536];
    static{
	//init BITSPERBYTE
	for(int i=0;i<BITSPERSHORT.length;i++){
	    int c = 0;
	    int b = i;
	    while(b!=0){
		if((b&1l)!=0)
		    c++;
		b>>>=1;
	    }
	    BITSPERSHORT[i] = c;
	}
    }

    public VierGewinntTurn(){
	
    }

    VierGewinntTurn(long data1, long data2){
	firstPlayerChips = data1;
	secondPlayerChips = data2;
    }

    VierGewinntTurn(long[] data){
	this(data[0],data[1]);
    }


    public VierGewinntTurn(VierGewinntTurn lastTurn){
	firstPlayerChips = lastTurn.firstPlayerChips;
	secondPlayerChips = lastTurn.secondPlayerChips;
    }

    public VierGewinntTurn(VierGewinntTurn lastTurn, int player, int column){
	this(lastTurn);
	firstPlayerChips = lastTurn.firstPlayerChips;
	secondPlayerChips = lastTurn.secondPlayerChips;

	if(!VP.SPEEDOPTIMIZED){
	    if(column<0||column>7)
		throw new IllegalArgumentException(
					"Illegal value for column: "+column);

	    if(player!=FIRSTPLAYER && player!=SECONDPLAYER)
		throw new IllegalArgumentException(
					 "Illegal value for player: "+player);

	}	
	long newChipMask = 0x8000000000000000l>>>column;
	long lastValidChipMask = 0;


	long chips = firstPlayerChips | secondPlayerChips;

	for(int i=0;i<8;i++,newChipMask>>>=8){
	    if((chips&newChipMask)==0)
		lastValidChipMask = newChipMask;
	     else break;
	}

	if(!VP.SPEEDOPTIMIZED){
	    if(lastValidChipMask==0)
		throw new IllegalArgumentException(
					 "No more space in column: "+column);
	}

	if(player==FIRSTPLAYER)
	    firstPlayerChips |= lastValidChipMask;
	 else secondPlayerChips |= lastValidChipMask;

	if(VP.USESYMMETRY)
	    makeSymmetricUnique();
    }

    long[] getInternalData(){
	return new long[]{firstPlayerChips,secondPlayerChips};
    }
    public int countChips(){
	final long INTMASK=0xffffffffl;
	final int SHORTMASK=0xffff;
	int chipsCount;

	int pi = (int) (firstPlayerChips&INTMASK);
	chipsCount = BITSPERSHORT[pi&SHORTMASK];
	chipsCount += BITSPERSHORT[(pi>>>16)&SHORTMASK];

	pi = (int) ((firstPlayerChips>>>32)&INTMASK);
	chipsCount += BITSPERSHORT[pi&SHORTMASK];
	chipsCount += BITSPERSHORT[(pi>>>16)&SHORTMASK];

	pi = (int) (secondPlayerChips&INTMASK);
	chipsCount += BITSPERSHORT[pi&SHORTMASK];
	chipsCount += BITSPERSHORT[(pi>>>16)&SHORTMASK];

	pi = (int) ((secondPlayerChips>>>32)&INTMASK);
	chipsCount += BITSPERSHORT[pi&SHORTMASK];
	chipsCount += BITSPERSHORT[(pi>>>16)&SHORTMASK];

	return chipsCount;
    }

    /**
       Every playfield has a symmetric weight of -1 or 0
       after this method call.
       @see calculateSymmetricWeight
     */
    private void makeSymmetricUnique(){
	int sw = calculateSymmetricWeight(firstPlayerChips);
	if(0 == sw)
	    sw = calculateSymmetricWeight(secondPlayerChips);
	if(sw>0)
	    mirrorField();
    }

    /**
       Calculates the symmetric weight of the playfield.
       That means the columns are compared. First column 0 and 7,
       then column 1 and 6, and so on, until two unequal columns
       are found. Then it returns -1, if the left column is higher
       or +1 if the right column is higher.
       If no unequal column is found, the function returns 0.
       @param chips The playfieldchips to compare
       @return The symmetric weight
     */
    private static int calculateSymmetricWeight(long chips){
	for(int i=0;i<(VP.MAXCOLUMNS+1)/2;i++){
	    int ls = 7-i; //left shift
	    int rs = i+(8-VP.MAXCOLUMNS); //right shift

	    long colr = ((chips>>>rs)&COLMASK);
	    long coll = ((chips>>>ls)&COLMASK);

	    if(coll!=colr)
		return (colr>coll)?1:-1;
	}

	return 0;

    }

    /**
       The left columns are right afterwards and vice versa.
    */
    private void mirrorField(){
	firstPlayerChips = mirrorChips(firstPlayerChips);
	secondPlayerChips = mirrorChips(secondPlayerChips);
    }

    /**
       The left columns are right afterwards and vice versa.
       @param chips The chips to mirrror
       @return The mirrored chips.
    */
    private static long mirrorChips(long chips){
	long ret = 0;

	for(int i=0;i<VP.MAXCOLUMNS;i++){
	    int ls = i+(8-VP.MAXCOLUMNS); //left shift
	    int rs = 7-i; //right shift

	    ret |= ((chips>>>rs)&COLMASK)<<ls;
	}

	return ret;
    }

    public boolean equals(Object o){
	if(!VP.SPEEDOPTIMIZED){
	    if(!(o instanceof VierGewinntTurn))
		return false;
	}

	VierGewinntTurn t = (VierGewinntTurn)o;
	return  t.firstPlayerChips==firstPlayerChips &&
	    t.secondPlayerChips==secondPlayerChips;
    }

    public boolean columnHasSpace(int column){
	long completeField = firstPlayerChips|secondPlayerChips;
	return (completeField&(0x8000000000000000l>>>column
			       >>>(8*(8-VP.MAXROWS))))==0;
	//	return (completeField&(0x8000000000000000l>>>column))==0;
	
    }

    public String toString(){
	StringBuffer out = 
	    new StringBuffer("First Player: x\nSecond Player: o\n");
	long mask = 0x8000000000000000l;
	for(int i=0;i<VP.MAXCOLUMNS+2;i++)
	    out.append('+');
	out.append('\n');
	    
	mask>>>=(8*(8-VP.MAXROWS));
	for(int i=0;i<VP.MAXROWS;i++){
	    out.append('+');
	    for(int j=0;j<VP.MAXCOLUMNS;j++){
		if((firstPlayerChips&mask)!=0)
		    out.append('x');
		 else if((secondPlayerChips&mask)!=0)
		    out.append('o');
		 else out.append(' ');
		mask>>>=1;
	    }
	    mask>>>=(8-VP.MAXCOLUMNS);
	    out.append("+\n");
	}

	for(int i=0;i<VP.MAXCOLUMNS+2;i++)
	    out.append('+');
	out.append('\n');

	switch(getWinner()){
	case FIRSTPLAYER: out.append("First player wins!");break;
	case SECONDPLAYER: out.append("Second Player wins!");break;
	case DRAW: out.append("Draw");break;
	case NOBODY: out.append("Nobody has won yet");break;
	default: throw new RuntimeException(
			   "getWinner() returned wrong result: "+getWinner());
	    

	}
	out.append('\n');

	out.append("x:"+Long.toBinaryString(firstPlayerChips)+"\n");
	out.append("o:"+Long.toBinaryString(secondPlayerChips)+"\n");
	return out.toString();
    }

    public int getWinner(){
	long chips;
	int player;
	int count = countChips();
	
	if(count==VP.MAXCOLUMNS*VP.MAXROWS)
	    return DRAW;

	if((count%2)==0){
	    chips = secondPlayerChips;
	    player = SECONDPLAYER;
	}else{ 
	    chips = firstPlayerChips;
	    player = FIRSTPLAYER;
	}
	
 	if(rowsHaveSolution(chips))
	    return player;
  	if(columnsHaveSolution(chips))
	    return player;
  	if(diagonalsHaveSolution(chips))
	    return player;
    

	return NOBODY;
    }

    private static boolean rowsHaveSolution(long chips){
	final long MASK = 0xfl;
	final int MAXSOLUTIONCOLS = VP.MAXCOLUMNS-4+1;
	//final int MAXSOLUTIONCOLS = 5;

	chips>>>=8-VP.MAXCOLUMNS; //skip empty columns
	while(chips!=0){
	    for(int i=0;i<MAXSOLUTIONCOLS;i++){
		if((chips&MASK)==MASK)
		    return true;
		chips>>>=1;
	    }
	    chips>>>=8-MAXSOLUTIONCOLS; //move pattern to next row
	}
	return false;
    }

    /*
    private boolean columnsHaveSolution(long chips){
	final long MASK = 0x01010101l;
	while(chips!=0){
	    if((chips&MASK)==MASK)
		return true;
	    chips>>>=1;
	}
	return false;
    }
*/
    private static boolean columnsHaveSolution(long chips){
	final int MASK = 0x01010101;
 	final int MAXSOLUTIONROWS = VP.MAXROWS-4+1;
	final int SKIPCOLS = 8-VP.MAXCOLUMNS;

	for(int r=0;r<MAXSOLUTIONROWS;r++){
	    for(int c=SKIPCOLS;c<8;c++){
		int t = (int)(chips>>>((r*8)+c));
		if(t==0)
		    return false;
		if((t&MASK)==MASK)
		    return true;
	    }
	}
	return false;
    }

    private static boolean diagonalsHaveSolution(long chips){
	final int MAXSOLUTIONCOLS = VP.MAXCOLUMNS-4+1;
 	final int MAXSOLUTIONROWS = VP.MAXROWS-4+1;
	final int SKIPCOLS = 8-VP.MAXCOLUMNS;
	final int MASK1 = 0x08040201;
	final int MASK2 = 0x01020408;

	for(int r=0;r<MAXSOLUTIONROWS;r++){
	    for(int c=SKIPCOLS;c<8;c++){
		int t = (int)(chips>>>((r*8)+c));
		if(t==0)
		    return false;
		if((t&MASK1)==MASK1 || (t&MASK2)==MASK2)
		    return true;
	    }
	}
	return false;
    }

    /*
    private static boolean rowsHaveSolution(long chips){
	final long MASK = 0xfl;
	final int MAXSOLUTIONCOLS = VP.MAXCOLUMNS-4+1;

	while(chips!=0){
	    for(int i=0;i<MAXSOLUTIONCOLS;i++){
		if((chips&MASK)==MASK)
		    return true;
		chips>>>=1;
	    }
	    chips>>>=8-MAXSOLUTIONCOLS+1; //move pattern to next row
	}
	return false;
    }

    private static boolean columnsHaveSolution(long chips){
	final int MAXSOLUTIONROWS = VP.MAXROWS-4+1;
	final int MAXCOUNT = MAXSOLUTIONROWS*VP.MAXCOLUMNS;

	final int MASK = 0x01010101;
	int i=0;
	int c;
	do{
	    c = (int)(chips>>>(i++));
	    if((c&MASK)==MASK)
		return true;
	}while(c!=0 && (true || MAXCOUNT==64 | i<MAXCOUNT));

	return false;
    }

    private static boolean diagonalsHaveSolution(long chips){
	long chips2 = chips;
	final int MAXSOLUTIONCOLS = VP.MAXCOLUMNS-4+1;
	final long mask = 0x08040201l;
	while(chips!=0){
	    for(int i=0;i<MAXSOLUTIONCOLS;i++){
		if((chips&mask)==mask)
		    return true;
		chips>>>=1;
	    }
	    chips>>>=8-MAXSOLUTIONCOLS+1;
	}

	final long mask2 = 0x01020408l;
	while(chips2!=0){
	    for(int i=0;i<MAXSOLUTIONCOLS;i++){
		if((chips2&mask2)==mask2)
		    return true;
		chips2>>>=1;
	    }
	    chips2>>>=8-MAXSOLUTIONCOLS+1;
	}

	return false;
    }
    */
    /*
    private boolean rowsHaveSolution(long chips){
	final long MASK = 0xfl;
	while(chips!=0){
	    for(int i=0;i<5;i++){
		if((chips&MASK)==MASK)
		    return true;
		chips>>>=1;
	    }
	    chips>>>=3; //move pattern to next row
	}
	return false;
    }

    private boolean columnsHaveSolution(long chips){
	final long MASK = 0x01010101l;
	while(chips!=0){
	    if((chips&MASK)==MASK)
		return true;
	    chips>>>=1;
	}
	return false;
    }

    private boolean diagonalsHaveSolution(long chips){
	long mask = 0x08040201l;
	long chips2 = chips;
	while(chips!=0){
	    for(int i=0;i<5;i++){
		if((chips&mask)==mask)
		    return true;
		chips>>>=1;
	    }
	    chips>>>=3;
	}

	mask = 0x01020408l;
	while(chips2!=0){
	    for(int i=0;i<5;i++){
		if((chips2&mask)==mask)
		    return true;
		chips2>>>=1;
	    }
	    chips2>>>=3;
	}

	return false;
    }
*/
    private boolean isDraw(){
	//checks if every field contains a chip
	return countChips()==VP.MAXCOLUMNS*VP.MAXROWS;
    }

    public int hashCode(){
	long x = (firstPlayerChips*(secondPlayerChips^0xaaaaaaaaaaaaaaaal));
	while(x!=0 && (x&1l)==0)
	    x>>>=1;
	x>>>=1; //needed because the LSB is now always 1
	return (int)x;
	/*
	final int p1 = (int)firstPlayerChips;
	final int p2 = (int)secondPlayerChips;
	return (p1*(p2^0xaaaaaaaa))>>>(8-VP.MAXCOLUMNS+4);
	//return (int)((firstPlayerChips>>>32)^secondPlayerChips);
	*/
    }

}






